/*
 * netdrv.c  -  Handle different network drivers
 *
 * Copyright (C) 1998 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define NEED_BINARY 1
#include <common.h>
#include <nblib.h>
#include "makerom.h"



/*
 * Definitions local for this module
 */
#define COMSIG		0x4B47		/* Signature for DOS program header */
#define EXESIG		0x5A4D		/* Signature of DOS EXE files */

#define MAXARGSIZE	126		/* Max length of DOS argument line */
#define MAXPROGSIZE	65536L		/* Max size of DOS program */
#define MAXUNDISIZE	65536L		/* Max size of UNDI driver */
#define MAXPISIZE	16424L		/* Max size of PROTOCOL.INI image */

#define PSPPARA		(256L / 16L)	/* Size of DOS PSP */
#define DOSBASEPARA	6L		/* DOS base paragraphs required */

#define EXESECTSIZE	512		/* Size of one sector in EXE file */
#define EXESIGOFS	0x0000		/* Offset to EXE file signature */
#define EXELASTSECT	0x0002		/* Number of bytes in last EXE sector */
#define EXEFSIZE	0x0004		/* Number of sectors in EXE file */
#define EXEHEADSIZE	0x0008		/* Offset to size of EXE file header */
#define EXEMINPARA	0x000A		/* Required number of paragraphs */
#define EXEHDRSIZE	0x001A		/* Minimum size of EXE header */

#define UNDISIGNATURE	"UNDI"		/* UNDI file signature */
#define UNDISIGOFS	0x0000		/* Offset to UNDI file signature */
#define UNDITEXT	0x0004		/* Size of text segment in bytes */
#define UNDIDATA	0x0006		/* Size of data segment in bytes */
#define UNDIBSS		0x0008		/* Size of BSS segment in bytes */
#define UNDIHDRSIZE	0x000E		/* Size of UNDI file header */



/*
 * Variables local to this module
 */
static long dosminpara;			/* Minimum no. of DOS paragraphs */
static long dosmaxpara;			/* Maximum no. of DOS paragraphs */



/*
 * Copy DOS program into output file
 */
static unsigned long copy_prog(fname, args, minsize, maxsize, outfile)
char *fname;
char *args;
long minsize;
long maxsize;
int outfile;
{
  struct stat sbuf;
  __u8 inbuf[BLKSIZE];
  __u8 firstbuf[BLKSIZE];
  unsigned long progsize;
  unsigned long writecnt = 0L;
  unsigned int inbuflen, firstbuflen;
  int i, len, infile;
  __u16 *u16p;

  /*
   * Each DOS program requires a header which looks like:
   *
   *	Offset
   *	 0000  -  Two byte magic cookie
   *	 0002  -  Length of command line in bytes
   *	 0003  -  Command line (not terminated by zero)
   *	 var   -  Length of program image in bytes
   *	 var+2 -  Amount of memory to allocate in paragraphs
   *	 var+4 -  Program image
   *
   * Setup everything up to the command line.
   */
  memset(firstbuf, 0, sizeof(firstbuf));
  firstbuf[0] = (__u8)(COMSIG & 0xff);
  firstbuf[1] = (__u8)((COMSIG >> 8) & 0xff);

  if (args != NULL) {
	if ((len = strlen(args)) > MAXARGSIZE - 1) {
		prnerr0("DOS program argument too large");
		exit(EXIT_MAKEROM_ARGSIZE);
	}
	firstbuf[2] = (__u8)((len + 1) & 0xff);
	firstbuf[3] = (__u8)0x20;	/* cmd line has to start with a blank */
	for (firstbuflen = 4, i = 0;
	     firstbuflen < sizeof(firstbuf) && i < len && args[i];
	     firstbuflen++, i++)
		firstbuf[firstbuflen] = (__u8)(args[i] & 0xff);
  } else {
	firstbuf[2] = 0;
	firstbuflen = 3;
  }

  /* Now open the DOS program image file and determine it's size. */
  if ((infile = open(fname, O_RDONLY | O_BINARY)) < 0) {
	prnerr1("unable to open DOS program %s", fname);
	exit(EXIT_MAKEROM_OPENDRV);
  }
  if (fstat(infile, &sbuf) < 0) {
	prnerr1("unable to stat DOS program %s", fname);
	exit(EXIT_STAT);
  }
  if (sbuf.st_size > MAXPROGSIZE) {
	prnerr1("DOS program %s too large", fname);
	exit(EXIT_MAKEROM_DOSSIZE);
  }
  progsize = sbuf.st_size;
  u16p = (__u16 *)(&firstbuf[firstbuflen]);
  assign(*u16p, htot(low_word(progsize)));
  firstbuflen += 2;

  /*
   * Read first sector of DOS program image and check it's type. In case of
   * a COM program, we set the number of required memory paragraphs to the
   * total size of the program. Otherwise, for an EXE program we can take
   * this value out of the EXE file header.
   */
  inbuflen = nbread((unsigned char *)inbuf, sizeof(inbuf), infile);
  if (ttoh(getval(*((__u16 *)(&inbuf[EXESIGOFS])))) == EXESIG) {
	unsigned long exehlen = 0L;
	unsigned long exesize, l;

	if (inbuflen >= EXEHDRSIZE)
		exehlen = ttoh(getval(*((__u16 *)(&inbuf[EXEHEADSIZE])))) * 16;
	if (exehlen < EXEHDRSIZE) {
		prnerr1("invalid EXE header in file %s", fname);
		exit(EXIT_MAKEROM_INVEXE);
	}
	l = ttoh(getval(*((__u16 *)(&inbuf[EXELASTSECT]))));
	exesize  = (ttoh(getval(*((__u16 *)(&inbuf[EXEFSIZE])))) - 1) * EXESECTSIZE;
	exesize += ttoh(getval(*((__u16 *)(&inbuf[EXEMINPARA])))) * 16;
	exesize += (l == 0L ? EXESECTSIZE : l);
	if (exesize > MAXPROGSIZE) {
		prnerr1("program %s requires too much memory", fname);
		exit(EXIT_MAKEROM_INVEXE);
	}
	if (maxsize < 0L)
		maxsize = ((exesize + 15L) / 16L + PSPPARA) * 16L;
	u16p = (__u16 *)(&firstbuf[firstbuflen]);
	assign(*u16p, htot(low_word((exesize + 15L) / 16L)));
	firstbuflen += 2;
  } else {
	unsigned long progpara;

	progpara = ((progsize + 15L) / 16L) + PSPPARA;
	if (maxsize < 0L)
		maxsize = (progpara * 16L + 0xFFFFL) & ~0xFFFFL;
	u16p = (__u16 *)(&firstbuf[firstbuflen]);
	assign(*u16p, htot(low_word(progpara)));
	firstbuflen += 2;
  }

  /* Determine the amount of memory required for the DOS simulator */
  if (minsize < 0L)
	minsize = maxsize;
  if (minsize > maxsize)
	maxsize = minsize;
  dosmaxpara = dosminpara + ((maxsize + 15L) / 16L) + 1;
  dosminpara = dosminpara + ((minsize + 15L) / 16L) + 1;

  /* Now first write the program header and then the program image */
  writecnt += nbwrite(firstbuf, firstbuflen, outfile);
  while (TRUE) {
	writecnt += nbwrite(inbuf, inbuflen, outfile);
	if ((inbuflen = nbread(inbuf, BLKSIZE, infile)) == 0)
		break;
  }
  close(infile);
  return(writecnt);
}



/*
 * Write packet driver programs into output file
 */
static unsigned long dopktdrv(np, parareq, outfile)
struct netdrvdef *np;
unsigned long *parareq;
int outfile;
{
  struct pktdrvdef *pdp = &(np->driverdefs.pd);
  struct i_long il;
  unsigned long writecnt = 0L;
  int i;

  /* Setup minimum amount of paragraphs required for DOS simulator */
  dosminpara = dosmaxpara = DOSBASEPARA;

  /* Write the number of DOS programs into output file */
  assign(il.low, htot(low_word(pdp->prognum)));
  writecnt += nbwrite((__u8 *)&(il.low), sizeof(il.low), outfile);

  /* Write all DOS programs into output file */
  for (i = 0; i < pdp->prognum; i++) {
	writecnt += copy_prog(pdp->prognames[i], pdp->progargs[i],
				pdp->minsizes[i], pdp->maxsizes[i], outfile);
	if (verbose > 2)
		printf("End offset of prog %d:           %ld\n", i, writecnt);
  }
  *parareq += dosmaxpara;
  return(writecnt);
}



/*
 * Write NDIS driver programs into output file
 */
static unsigned long dondisdrv(np, parareq, outfile)
struct netdrvdef *np;
unsigned long *parareq;
int outfile;
{
  struct ndisdef *ndp = &(np->driverdefs.ndis);
  struct i_long il;
  unsigned long writecnt = 0L;
  int i;

  /* Write the size and image of protocol.ini into the output file */
  if (ndp->protinisize > MAXPISIZE) {
	prnerr0("PROTOCOL.INI image too large");
	exit(EXIT_MAKEROM_PISIZE);
  }
  assign(il.low, htot(low_word(ndp->protinisize)));
  writecnt += nbwrite((__u8 *)&(il.low), sizeof(il.low), outfile);
  writecnt += nbwrite((__u8 *)(ndp->protocolini), ndp->protinisize, outfile);

  /* Setup minimum amount of paragraphs required for DOS simulator */
  dosminpara = dosmaxpara = DOSBASEPARA + ((ndp->protinisize + 15L) / 16L) + 1;

  /* Next write the number of DOS programs into output file */
  assign(il.low, htot(low_word(ndp->prognum)));
  writecnt += nbwrite((__u8 *)&(il.low), sizeof(il.low), outfile);

  /* Write all DOS programs into output file */
  for (i = 0; i < ndp->prognum; i++) {
	writecnt += copy_prog(ndp->prognames[i], ndp->progargs[i],
				ndp->minsizes[i], ndp->maxsizes[i], outfile);
	if (verbose > 2)
		printf("End offset of prog %d:           %ld\n", i, writecnt);
  }
  *parareq += dosmaxpara;
  return(writecnt);
}



/*
 * Write UNDI driver into output file
 */
static unsigned long doundidrv(np, parareq, outfile)
struct netdrvdef *np;
unsigned long *parareq;
int outfile;
{
  struct undidef *up = &(np->driverdefs.undi);
  struct stat sbuf;
  unsigned long undisize;
  unsigned long writecnt = 0L;
  __u8 inbuf[BLKSIZE];
  int infile, inbuflen;


  /* Open the UNDI driver image file */
  if ((infile = open(up->name, O_RDONLY | O_BINARY)) < 0) {
	prnerr1("unable to open UNDI driver%s", up->name);
	exit(EXIT_MAKEROM_OPENDRV);
  }
  if (fstat(infile, &sbuf) < 0) {
	prnerr1("unable to stat UNDI driver %s", up->name);
	exit(EXIT_STAT);
  }
  if (sbuf.st_size > MAXUNDISIZE) {
	prnerr1("UNDI driver %s too large", up->name);
	exit(EXIT_MAKEROM_UNDISIZE);
  }

  /*
   * Read the first sector of the UNDI driver and check that it has a
   * correct header.
   */
  inbuflen = nbread((__u8 *)inbuf, BLKSIZE, infile);
  if (inbuflen < UNDIHDRSIZE ||
      !bytecmp(UNDISIGNATURE, inbuf, strlen(UNDISIGNATURE))) {
	prnerr1("invalid UNDI driver header in file %s", up->name);
	exit(EXIT_MAKEROM_INVDRV);
  }
  undisize  = ttoh(getval(*((__u16 *)(&inbuf[UNDITEXT]))));
  undisize += ttoh(getval(*((__u16 *)(&inbuf[UNDIBSS]))));
  if (undisize > MAXUNDISIZE) {
	prnerr1("UNDI driver %s too large", up->name);
	exit(EXIT_MAKEROM_UNDISIZE);
  }
  *parareq += ((undisize + 15L) / 16L) + 1;

  /* Write the driver image into the output file */
  while (inbuflen > 0) {
	writecnt += nbwrite((__u8 *)inbuf, inbuflen, outfile);
	inbuflen = nbread((__u8 *)inbuf, BLKSIZE, infile);
  }
  close(infile);
  return(writecnt);
}



/*
 * Copy the network driver into the output file
 */
unsigned long donetdrv(np, parareq, outfile)
struct netdrvdef *np;
unsigned long *parareq;
int outfile;
{
  switch (np->drivertype) {
	case DRVTYPE_PD:
		return(dopktdrv(np, parareq, outfile));
	case DRVTYPE_NDIS:
		return(dondisdrv(np, parareq, outfile));
	case DRVTYPE_UNDI:
		return(doundidrv(np, parareq, outfile));
  }
  return(0);
}

